"""
Defines the RecoverpyApp class which serves as the main app orchestrator for the application.
"""
from typing import Dict, cast

from textual.app import App
from textual.screen import Screen

from recoverpy.lib.env_check import verify_app_environment
from recoverpy.log.logger import log
from recoverpy.ui.css import get_css
from recoverpy.ui.screens.modal import install_and_push_modal
from recoverpy.ui.screens.screen_params import ParamsScreen
from recoverpy.ui.screens.screen_path_edit import PathEditScreen
from recoverpy.ui.screens.screen_result import ResultScreen
from recoverpy.ui.screens.screen_save import SaveScreen
from recoverpy.ui.screens.screen_search import SearchScreen


class RecoverpyApp(App[None]):
    screens: Dict[str, Screen[None]]
    CSS_PATH = get_css()  # type: ignore[assignment]

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self._is_user_root = False
        self._initialize_screens()
        log.debug("Recoverpy app initialized")

    def _initialize_screens(self) -> None:
        self.screens = {
            "params": ParamsScreen(name="params"),
            "search": SearchScreen(name="search"),
            "result": ResultScreen(name="result"),
            "save": SaveScreen(name="save"),
            "path_edit": PathEditScreen(name="path_edit"),
        }
        for screen_name, screen_instance in self.screens.items():
            self.install_screen(screen_instance, screen_name)
            log.debug(f"Installed screen {screen_name}")

    async def on_mount(self) -> None:
        await self.push_screen("params")
        await verify_app_environment(self)

    async def on_params_screen_continue(self, message: ParamsScreen.Continue) -> None:
        log.info("User clicked continue on parameters screen")
        self.pop_screen()
        await self.push_screen("search")
        self.get_screen("search").post_message(
            SearchScreen.Start(message.searched_string, message.selected_partition)
        )

    async def on_search_screen_open(self, message: SearchScreen.Open) -> None:
        log.info("User clicked open on search screen")
        await self.push_screen("result")
        cast(ResultScreen, self.get_screen("result")).set(
            message.partition, message.block_size, message.inode
        )

    async def on_save_screen_saved(self, message: SaveScreen.Saved) -> None:
        log.info("User clicked save on save screen")
        await self._display_save_modal(message.save_path)

    async def _display_save_modal(self, save_path: str) -> None:
        await install_and_push_modal(
            self, "save-modal", f"Saved results to {save_path}"
        )



import argparse
import logging
from datetime import datetime
from os import path
from tempfile import gettempdir

from recoverpy.log.logger import log
from recoverpy.ui.app import RecoverpyApp


def _set_logger() -> None:
    global log

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", action="store_true", help="Enable logging")

    args = parser.parse_args()
    timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
    log_file_path = path.join(gettempdir(), f"recoverpy-{timestamp}.log")

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.CRITICAL,
        format="%(asctime)s - %(levelname)s - [%(threadName)s] - %(message)s",
        handlers=[logging.FileHandler(log_file_path)]
        if args.debug
        else [logging.NullHandler()],
    )
    log = logging.getLogger()


def main() -> None:
    _set_logger()
    log.info("Starting Recoverpy app")
    RecoverpyApp().run()


if __name__ == "__main__":
    main()



"""Screen displaying grep results."""

from asyncio import ensure_future, sleep
from typing import List

from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button, Label

from recoverpy.lib.search.search_engine import SearchEngine
from recoverpy.log.logger import log
from recoverpy.models.grep_result import GrepResult
from recoverpy.ui.widgets.grep_result_list import GrepResultList


class SearchScreen(Screen[None]):
    class Start(Message):
        def __init__(self, searched_string: str, selected_partition: str) -> None:
            super().__init__()
            self.searched_string = searched_string
            self.selected_partition = selected_partition

    class Open(Message):
        def __init__(self, inode: int, block_size: int, partition: str) -> None:
            super().__init__()
            self.inode = inode
            self.block_size = block_size
            self.partition = partition

    class InfoContainer(Horizontal):
        def __init__(self, *args, **kwargs) -> None:  # type: ignore
            super().__init__(classes="info-container", *args, **kwargs)

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        self.results: List[str] = []
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        self._grep_result_list = GrepResultList()
        self._result_count_label = Label("0", id="result-count")
        self._progress_title_label = Label("", id="progress-title")
        self._progress_label = Label("", id="progress")
        self._open_button = Button(label="Open", id="open-button", disabled=True)

        yield self._grep_result_list
        yield Vertical(
            self.InfoContainer(Label("- result count -", id="result-count-title")),
            self.InfoContainer(self._result_count_label),
            self.InfoContainer(self._progress_title_label),
            self.InfoContainer(self._progress_label),
            id="info-bar",
        )
        yield self._open_button
        yield Button("Exit", id="exit-button")
        log.debug("search - Search screen composed")

    async def on_search_screen_start(self, message: Start) -> None:
        self.search_engine = SearchEngine(
            message.selected_partition, message.searched_string
        )
        await self._wait_for_grep_list_focus()
        await self._start_search_engine()

    async def _wait_for_grep_list_focus(self) -> None:
        while self._grep_result_list not in self.focus_chain:
            continue

    async def _start_search_engine(self) -> None:
        await self.search_engine.start_search()
        ensure_future(
            self._grep_result_list.start_consumer(self.search_engine.list_items_queue)
        )
        ensure_future(self._update_progress_labels())

    async def _update_progress_labels(self) -> None:
        while True:
            self._result_count_label.update(
                str(self.search_engine.search_progress.result_count)
            )
            self._update_progress_percent_title()

            await sleep(0.1)

    def _update_progress_percent_title(self) -> None:
        if self.search_engine.search_progress.progress_percent == 0.0:
            return

        if self._progress_title_label.renderable == "":
            self._progress_title_label.update("- progress -")

        self._progress_label.update(
            f"{self.search_engine.search_progress.progress_percent}%"
        )

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        actions = {
            "exit-button": self._handle_exit_button,
            "open-button": self._handle_open_button,
        }

        button_id = event.button.id
        if button_id in actions:
            await actions[button_id]()

    async def _handle_exit_button(self) -> None:
        log.info("search - User clicked exit button")
        self.search_engine.stop_search()
        self.app.exit()
        exit()

    async def _handle_open_button(self) -> None:
        log.info("search - User clicked open button")
        selected_grep_result = self._get_selected_grep_result()
        log.info(f"search - Opening inode {selected_grep_result.inode}")
        self.app.post_message(
            self.Open(
                selected_grep_result.inode,
                self.search_engine.search_params.block_size,
                self.search_engine.search_params.partition,
            )
        )

    def _get_selected_grep_result(self) -> GrepResult:
        return self._grep_result_list.grep_results[self._grep_result_list.get_index()]

    async def on_list_view_highlighted(self) -> None:
        self._open_button.disabled = False



from dataclasses import dataclass
from typing import Optional


@dataclass
class Partition:
    name: str
    fs_type: str
    is_mounted: bool
    mount_point: Optional[str]

    def get_full_name(self) -> str:
        return f"/dev/{self.name}"



"""Screen used to select result save path."""

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.events import Event
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button

from recoverpy.log.logger import log
from recoverpy.ui.widgets.directory_tree import DirectoryTree


class PathEditScreen(Screen[None]):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self._directory_tree = DirectoryTree("/")

    class Confirm(Message):
        def __init__(self, selected_dir: str) -> None:
            super().__init__()
            self.selected_dir = selected_dir

    def compose(self) -> ComposeResult:
        yield self._directory_tree
        yield Horizontal(
            Button("Confirm", id="confirm-button"), id="path-edit-button-container"
        )
        log.info("path_edit - Path edit screen composed")

    async def on_button_pressed(self, event: Event) -> None:
        log.info("path_edit - Confirm button pressed")
        event.stop()
        self.app.get_screen("save").post_message(
            self.Confirm(self._directory_tree.selected_dir)
        )
        self.app.pop_screen()



from typing import Optional

from textual.widgets import Label, ListItem

from recoverpy.lib.helper import get_inode, get_printable
from recoverpy.log.logger import log


class GrepResult:
    def __init__(self, line: str):
        self.inode = get_inode(line)
        self.line = get_printable(line)
        self.list_item: Optional[ListItem] = None
        self.css_class = "grep-result"

    def create_list_item(self) -> None:
        try:
            self.list_item = ListItem(
                Label(str(self.line), markup=False),
                classes=self.css_class,
                id=f"grep-result-{self.inode}",
            )
        except RuntimeError as e:
            log.error(f"grep_result - Error creating list item: {e}")
            return



"""Screen displaying dd results."""
from subprocess import CalledProcessError
from typing import Optional, cast

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.screen import Screen
from textual.widgets import Button, Label, RichLog

from recoverpy.lib.helper import decode_result, get_dd_output, get_printable
from recoverpy.lib.saver import Saver
from recoverpy.log.logger import log
from recoverpy.ui.screens.screen_save import SaveScreen


class ResultScreen(Screen[None]):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self._saver = Saver()
        self._partition = ""
        self._block_size = 0
        self._inode = 0
        self._inode_label = Label("", id="inode-label")
        self._block_count_label = Label("0 block selected", id="block-count")
        self._block_content = RichLog(markup=False, wrap=True)
        self._raw_block_content: Optional[str] = None
        self._save_button = Button(label="Save", id="save-button", disabled=True)

    def compose(self) -> ComposeResult:
        yield Horizontal(self._inode_label, id="inode-label-container")
        yield Horizontal(self._block_content, id="block-content-container")
        yield Horizontal(
            self._block_count_label,
            id="block-count-container",
        )
        yield Horizontal(
            Button("Go back", id="go-back-button"), id="go-back-button-container"
        )
        yield Horizontal(
            Button("Previous", id="previous-button"),
            Button("Add block", id="add-block-button"),
            Button("Next", id="next-button"),
            id="block-buttons-container",
        )
        yield Horizontal(self._save_button, id="save-button-container")
        log.debug("result - Result screen composed")

    def set(self, partition: str, block_size: int, inode: int) -> None:
        self._saver.reset_results()
        self._save_button.disabled = True
        self._partition = partition
        self._block_size = block_size
        self._inode = inode
        self._update_ui()
        self._block_count_label.update("0 block selected")

    def _update_ui(self) -> None:
        self._update_inode_label()
        self._update_block_content()

    def _update_inode_label(self) -> None:
        self._inode_label.update(f"Result for inode {self._inode}")

    def _update_block_content(self) -> None:
        self._block_content.clear()
        try:
            self._raw_block_content = decode_result(
                get_dd_output(self._partition, self._block_size, self._inode)
            )
            self._block_content.write(get_printable(self._raw_block_content))
            log.info("Block content updated")
        except CalledProcessError:
            log.error(f"Cannot read block {self._inode}")
            self._block_content.write(f"Cannot read block {self._inode}")

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        button_id = event.button.id
        actions = {
            "go-back-button": self._handle_go_back,
            "previous-button": self._handle_previous,
            "add-block-button": self._handle_add_block,
            "next-button": self._handle_next,
            "save-button": self._handle_save,
        }

        if button_id in actions:
            await actions[button_id]()

    async def _handle_go_back(self) -> None:
        log.info("Go back button pressed")
        self.app.pop_screen()

    async def _handle_previous(self) -> None:
        if self._inode > 0:
            log.info("Previous button pressed")
            self._inode -= 1
            self._update_ui()

    async def _handle_add_block(self) -> None:
        if self._raw_block_content:
            log.info("Add block button pressed")
            self._saver.add_result(self._inode, self._raw_block_content)
            count = self._saver.get_blocks_to_save_count()
            self._block_count_label.update(
                f"{count} block{'s' if count > 1 else ''} selected"
            )
            self._save_button.disabled = False

    async def _handle_next(self) -> None:
        log.info("Next button pressed")
        self._inode += 1
        self._update_ui()

    async def _handle_save(self) -> None:
        log.info("Save button pressed")
        cast(SaveScreen, self.app.get_screen("save")).set_saver(self._saver)
        await self.app.push_screen("save")



"""First screen displayed to the user.
Allows user to enter search string and select partition."""

from textual.app import ComposeResult
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button, Input, Label

from recoverpy.log.logger import log
from recoverpy.models.partition import Partition
from recoverpy.ui.widgets.partition_list import PartitionList


class ParamsScreen(Screen[None]):
    class Continue(Message):
        def __init__(self, searched_string: str, selected_partition: str) -> None:
            super().__init__()
            self.searched_string = searched_string
            self.selected_partition = selected_partition

    def compose(self) -> ComposeResult:
        self._partition_list = PartitionList()
        self._search_input = Input(
            name="search", id="search-input", placeholder="Search"
        )
        self._start_search_button = Button(
            label="Start search", id="start-search-button", disabled=True
        )

        yield Label("Type a text to search for:")
        yield self._search_input
        yield Label("Available partitions:")
        yield self._partition_list
        yield self._start_search_button
        log.debug("params - Parameters screen composed")

    async def on_button_pressed(self) -> None:
        highlighted_child = self._partition_list.highlighted_child
        searched_string = self._search_input.value.strip()

        if highlighted_child is None:
            log.warn("No partition selected for search")
            return

        if not searched_string:
            log.warn("No search string entered")
            return

        selected_partition: Partition = self._partition_list.list_items[
            highlighted_child.id
        ]

        log.info(
            f"User selected partition {selected_partition.get_full_name()} and search string `{searched_string}`"
        )

        self.app.post_message(
            self.Continue(searched_string, selected_partition.get_full_name())
        )

    async def on_input_changed(self, event: Input.Changed) -> None:
        self._start_search_button.disabled = len(event.value.strip()) == 0



from subprocess import check_output
from typing import List, Optional, Tuple

from recoverpy.models.partition import Partition

_IGNORED_PARTITIONS_TYPES: Tuple[str, str] = (" loop ", "swap")


_IGNORED_PARTITION_TYPES: Tuple[str, str] = (" loop ", "swap")


def get_partitions() -> List[Partition]:
    lsblk_output: str = _fetch_lsblk_output()
    return _parse_lsblk_output(lsblk_output)


def _fetch_lsblk_output() -> str:
    return check_output(
        ["lsblk", "-r", "-n", "-o", "NAME,TYPE,FSTYPE,MOUNTPOINT"],
        encoding="utf-8",
    )


def _parse_lsblk_output(lsblk_output: str) -> List[Partition]:
    partitions = [
        _parse_partition(line)
        for line in lsblk_output.splitlines()
        if not any(ignored in line for ignored in _IGNORED_PARTITION_TYPES)
    ]
    return [p for p in partitions if p]


def _parse_partition(line: str) -> Optional[Partition]:
    values = line.strip().split()

    if len(values) < 3:
        return None

    name, _, fs_type, *mount_info = values
    is_mounted = bool(mount_info)
    mount_point = mount_info[0] if mount_info else None

    return Partition(
        name=name,
        fs_type=fs_type,
        is_mounted=is_mounted,
        mount_point=mount_point,
    )



from __future__ import annotations

from queue import Queue
from typing import List

from recoverpy.lib.helper import decode_result, get_dd_output, get_inode
from recoverpy.log.logger import log
from recoverpy.models.search_params import SearchParams


class ResultProcessor:
    def __init__(self, search_params: SearchParams):
        self.search_params = search_params

    def get_new_results(self, queue_object: Queue[bytes]) -> List[str]:
        """Consume grep output queue and format results."""

        queue_list: List[bytes] = list(queue_object.queue)
        queue_size = len(queue_list)
        queue_object.queue.clear()
        if queue_size == 0:
            return []

        decoded_results: List[str] = [decode_result(r) for r in queue_list]

        if not self.search_params.is_multi_line:
            return decoded_results

        final_results = [
            result for result in decoded_results if self.is_result_format_valid(result)
        ]
        log.debug(f"result_processor - Found {len(final_results)} new results")
        return final_results

    def is_result_format_valid(self, result: str) -> bool:
        inode = int(get_inode(result))
        block_factor = self.search_params.block_size * 8

        both_block_output = self._get_combined_block_output(inode, block_factor)
        return all(
            line in both_block_output for line in self.search_params.searched_lines
        )

    def _get_combined_block_output(self, inode: int, block_factor: int) -> str:
        block_index = inode // block_factor
        block_output = get_dd_output(
            self.search_params.partition, block_factor, block_index
        )
        next_block_output = get_dd_output(
            self.search_params.partition, block_factor, block_index + 1
        )

        return decode_result(block_output) + decode_result(next_block_output)

    def fix_line_start(self, line: str) -> str:
        result_index: int = line.find(self.search_params.searched_lines[0])
        return line[min(result_index, 15) :]

    def fix_inode(self, inode: int) -> int:
        inode //= self.search_params.block_size

        for _ in range(10):
            dd_output = self._get_dd_output(inode)
            if self.search_params.searched_lines[0] in decode_result(dd_output):
                return inode
            inode += 1
        return inode

    def _get_dd_output(self, inode: int) -> bytes:
        return get_dd_output(
            self.search_params.partition, self.search_params.block_size, inode
        )



from __future__ import annotations

from io import BufferedReader
from multiprocessing import Queue

from recoverpy.log.logger import log


def enqueue_grep_output(out: BufferedReader, queue: Queue[bytes]) -> None:
    log.debug("grep_consumer - Grep output enqueue thread started")
    for line in iter(out.readline, b""):
        queue.put(line)
    out.close()



from re import findall
from subprocess import DEVNULL, check_output
from time import sleep
from typing import List

from recoverpy.models.search_progress import SearchProgress


def monitor_search_progress(grep_pid: int, progress: SearchProgress) -> None:
    while True:
        output: str = get_progress_output(grep_pid)

        if not output:
            progress.progress_percent = 100.0

        percent: List[str] = findall(r"(\d+\.\d+)%", output)

        if not percent:
            continue

        progress.progress_percent = float(percent[0])
        sleep(0.5)


def get_progress_output(grep_pid: int) -> str:
    return check_output(["progress", "-p", str(grep_pid)], stderr=DEVNULL).decode(
        "utf8"
    )



from __future__ import annotations

from asyncio import AbstractEventLoop
from asyncio import Queue as AsyncQueue
from asyncio import new_event_loop, set_event_loop
from queue import Queue
from subprocess import Popen
from time import sleep
from typing import List

from recoverpy.lib.search.result_processor import ResultProcessor
from recoverpy.lib.search.thread_factory import (
    start_grep_process,
    start_progress_monitoring_thread,
    start_result_dequeue_thread,
    start_result_enqueue_thread,
)
from recoverpy.log.logger import log
from recoverpy.models.grep_result import GrepResult
from recoverpy.models.search_params import SearchParams
from recoverpy.models.search_progress import SearchProgress


class SearchEngine:
    _grep_process: Popen[bytes]
    _seen_inodes: set[int] = set()

    def __init__(self, partition: str, searched_string: str):
        self._initialize_search_components(partition, searched_string)
        self.results_queue: Queue[bytes] = Queue()
        self.list_items_queue: AsyncQueue[GrepResult] = AsyncQueue()

    def _initialize_search_components(
        self, partition: str, searched_string: str
    ) -> None:
        self.search_params = SearchParams(partition, searched_string)
        self.search_progress = SearchProgress()
        self.result_processor = ResultProcessor(self.search_params)

    async def start_search(self) -> None:
        self._initialize_grep_process()
        self._start_auxiliary_threads()

    def stop_search(self) -> None:
        self._grep_process.kill()

    def _initialize_grep_process(self) -> None:
        self._grep_process = start_grep_process(
            searched_string=self.search_params.searched_lines[0],
            partition=self.search_params.partition,
        )

    def _start_auxiliary_threads(self) -> None:
        start_result_enqueue_thread(self._grep_process, self.results_queue)
        start_result_dequeue_thread(self._dequeue_results)
        start_progress_monitoring_thread(self._grep_process, self.search_progress)

    def _dequeue_results(self) -> None:
        loop = new_event_loop()
        while True:
            results = self.result_processor.get_new_results(self.results_queue)
            self._process_results(results, loop)
            log.debug(f"search_engine - Dequeued {len(results)} results")
            sleep(0.1)

    def _process_results(self, results: List[str], loop: AbstractEventLoop) -> None:
        for result in results:
            grep_result = self._create_grep_result(
                result, self.search_progress.result_count
            )

            if grep_result.inode not in self._seen_inodes:
                self._seen_inodes.add(grep_result.inode)
                loop.run_until_complete(self.list_items_queue.put(grep_result))
                self.search_progress.result_count += 1

    def _create_grep_result(self, result: str, result_index: int) -> GrepResult:
        grep_result = GrepResult(result)
        self._configure_grep_result(grep_result, result_index)
        log.debug(f"search_engine - Created grep result {grep_result.inode}")
        return grep_result

    def _configure_grep_result(
        self, grep_result: GrepResult, result_index: int
    ) -> None:
        grep_result.inode = self.result_processor.fix_inode(grep_result.inode)
        grep_result.line = self.result_processor.fix_line_start(grep_result.line)
        grep_result.css_class = (
            "grep-result-odd" if result_index % 2 == 0 else "grep-result-even"
        )



from typing import Callable, List, Optional

from textual.app import App, ComposeResult
from textual.containers import Grid, Horizontal
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Button, Label

from recoverpy.log.logger import log


class Modal(ModalScreen[None]):
    def __init__(self, name: str, message: str, callback: Optional[Callable] = None):  # type: ignore
        super().__init__(name=name)
        self._message_label = Label(message, id="modal-message")
        self._callback = callback or self.app.pop_screen
        self._is_error = "error" in name

    def compose(self) -> ComposeResult:
        grid_components: List[Widget] = [
            self._message_label,
            Horizontal(
                Button("OK", id="ok-button"),
                id="modal-button-container",
            ),
        ]

        if self._is_error:
            grid_components.insert(
                0, Label("An error occurred:", id="modal-error-label")
            )

        yield Grid(*grid_components, id="modal-container")
        log.info("Modal screen composed")

    def on_button_pressed(self) -> None:
        self._callback()


async def install_and_push_modal(
    app: App[None], name: str, message: str, callback: Optional[Callable] = None  # type: ignore
) -> None:
    modal = Modal(name, message=message, callback=callback)
    app.install_screen(modal, name)
    await app.push_screen(name)



from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional

from recoverpy.log.logger import log


class Saver:
    def __init__(self) -> None:
        self.save_path: Path = Path.cwd()
        self.last_saved_file: Optional[Path] = None
        self._results: Dict[int, str] = {}

    def add_result(self, inode: int, result: str) -> None:
        self._results[inode] = result
        log.info(f"Saver - Added result for inode {inode}, {len(self._results)} total")

    def reset_results(self) -> None:
        self._results.clear()
        log.info("Saver - Results reset")

    def save_results(self) -> None:
        ordered_blocks: List[str] = [
            self._results[key] for key in sorted(self._results)
        ]
        final_output: str = "\n".join(ordered_blocks)

        self._save_result_string(final_output)
        self.reset_results()

    def set_save_path(self, path: str) -> None:
        self.save_path = Path(path)
        log.info(f"Saver - Save path set to {self.save_path}")

    def get_blocks_to_save_count(self) -> int:
        return len(self._results)

    def _save_result_string(self, result: str) -> None:
        if not self.save_path:
            return

        timestamp = datetime.now().strftime("recoverpy-save-%Y-%m-%d-%H%M%S")
        file_name: Path = self.save_path / timestamp

        self._write_to_file(file_name, result)
        log.info(f"Saver - Saved result to {file_name}")

    def _write_to_file(self, file_name: Path, content: str) -> None:
        with open(file_name, "w") as save_file:
            save_file.write(content)

        self.last_saved_file = file_name



from __future__ import annotations

from queue import Queue
from subprocess import PIPE, Popen
from threading import Thread
from typing import Callable

from recoverpy.lib.helper import is_dependency_installed
from recoverpy.lib.search.grep_consumer import enqueue_grep_output
from recoverpy.lib.search.progress_monitoring import monitor_search_progress
from recoverpy.log.logger import log
from recoverpy.models.search_progress import SearchProgress


def start_grep_process(searched_string: str, partition: str) -> Popen[bytes]:
    log.debug(f"thread_factory - Starting grep process for {searched_string}")
    return Popen(
        ["grep", "-a", "-b", f"{searched_string}", partition],
        stdin=None,
        stdout=PIPE,
        stderr=None,
    )


def start_result_enqueue_thread(
    grep_process: Popen[bytes], queue: Queue[bytes]
) -> None:
    log.debug("thread_factory - Starting result enqueue thread")
    Thread(
        target=enqueue_grep_output,
        args=(grep_process.stdout, queue),
        daemon=True,
        name="enqueue-grep-output-thread",
    ).start()


def start_result_dequeue_thread(dequeue_results: Callable[[], None]) -> None:
    log.debug("thread_factory - Starting result dequeue thread")
    Thread(
        target=dequeue_results,
        daemon=True,
        name="dequeue-results-thread",
    ).start()


def start_progress_monitoring_thread(
    grep_process: Popen[bytes], progress: SearchProgress
) -> None:
    if not is_dependency_installed(command="progress"):
        log.debug("thread_factory - progress not installed, skipping progress thread")
        return

    log.debug("thread_factory - Starting progress monitoring thread")
    Thread(
        target=monitor_search_progress,
        args=(grep_process.pid, progress),
        daemon=True,
        name="progress-monitoring-thread",
    ).start()



from os import geteuid
from platform import system
from sys import exit, version_info

from textual.app import App

from recoverpy.lib.helper import is_dependency_installed
from recoverpy.log.logger import log
from recoverpy.ui.screens.modal import install_and_push_modal

_root_error_message = "The current user is not root, the search is likely to not work. Run this app as root or use sudo."
_version_error_message = (
    "The current Python version is not supported. Please use Python 3.8 or higher."
)
_linux_error_message = (
    "Your system may not be Linux-based, the application might not work correctly."
)
_dependencies_error_message = "Some dependencies are not installed. Please install grep, dd, lsblk and blockdev and restart the application."


async def verify_app_environment(app: App[None]) -> None:
    if not _is_version_supported():
        log.error("app - Python version not supported")
        await install_and_push_modal(
            app, "version-error-modal", _version_error_message, exit
        )
    if not _is_linux():
        log.warn("app - System is not Linux-based")
        await install_and_push_modal(app, "linux-error-modal", _linux_error_message)
    if not _is_user_root():
        log.warn("app - User is not root")
        await install_and_push_modal(app, "root-error-modal", _root_error_message)
    if not _are_system_dependencies_installed():
        log.warn("app - Some dependencies are not installed")
        await install_and_push_modal(
            app, "dependencies-error-modal", _dependencies_error_message
        )


def _is_user_root() -> bool:
    return geteuid() == 0


def _is_version_supported() -> bool:
    return version_info >= (3, 8)


def _is_linux() -> bool:
    return "linux" in system().lower()


def _are_system_dependencies_installed() -> bool:
    for dependency in ("grep", "dd", "lsblk", "blockdev"):
        if not is_dependency_installed(dependency):
            return False
    return True



"""Screen used to confirm save."""
from typing import Optional

from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button, Label

from recoverpy.lib.saver import Saver
from recoverpy.log.logger import log
from recoverpy.ui.screens.screen_path_edit import PathEditScreen


class SaveScreen(Screen[None]):
    class Saved(Message):
        def __init__(self, save_path: str) -> None:
            super().__init__()
            self.save_path = save_path

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self._saver: Optional[Saver] = None
        self._save_path_label = Label("", id="save-path-label")

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label("Current save path:", id="save-path-title-label"),
            self._save_path_label,
            id="save-path-container",
        )
        yield Horizontal(
            Button("Go back", id="go-back-button"),
            Button("Edit save path", id="edit-save-path-button"),
            Button("Save", id="save-button"),
            id="action-buttons-container",
        )
        log.debug("save - Save screen composed")

    def set_saver(self, saver: Saver) -> None:
        self._saver = saver
        self._update_save_path_label()

    def _update_save_path_label(self) -> None:
        if self._saver:
            self._save_path_label.update(str(self._saver.save_path))
            log.info(f"save - Save path set to {self._saver.save_path}")
        else:
            log.error("save - Saver not set for path setting")

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        actions = {
            "go-back-button": self._handle_go_back,
            "edit-save-path-button": self._handle_edit_save_path,
            "save-button": self._handle_save,
        }

        button_id = event.button.id
        if button_id in actions:
            await actions[button_id]()

    async def _handle_go_back(self) -> None:
        log.info("save - Go back button pressed")
        self.app.pop_screen()

    async def _handle_edit_save_path(self) -> None:
        log.info("save - Edit save path button pressed")
        await self.app.push_screen("path_edit")

    async def _handle_save(self) -> None:
        if self._saver:
            log.info("save - Save button pressed")
            self._saver.save_results()
            self.app.pop_screen()
            if self._saver.last_saved_file:
                self.app.post_message(
                    self.Saved(str(self._saver.save_path / self._saver.last_saved_file))
                )
        else:
            log.error("save - Saver not set for saving")

    async def on_path_edit_screen_confirm(self, event: PathEditScreen.Confirm) -> None:
        if self._saver:
            self._saver.set_save_path(event.selected_dir)
            self._update_save_path_label()
        else:
            log.error("save - Saver not set for path editing")
